from pathlib import Path

import click
from pymatgen.io.cif import CifParser
from pymatgen.io.vasp import Poscar

from . import cli
from ..potential import Potential


def get_structure(filename):
    from pymatgen.io.cif import CifParser
    from pymatgen.io.vasp import Poscar

    if filename.lower().endswith('.cif'):
        return CifParser(filename).get_structure()[0]
    elif filename.lower().endswith('poscar'):
        return Poscar.from_file(filename).structure
    else:
        raise ValueError('Cannot determine file type from filename [.cif, poscar]')


@cli.command()
@click.option('-s', '--structure', type=click.Path(exists=True), required=True)
@click.option('-p', '--potential', type=click.Path(exists=True), required=True)
@click.option('-d', '--directory', type=click.Path(exists=True), default=".")
@click.option('--software', type=click.Choice(['lammps']), default="lammps")
@click.option('-c', '--calculation', type=click.Choice(['relax', 'static']), default="relax")
@click.pass_context
def test(ctx, structure, potential, directory, software, calculation):
    directory = str(Path(directory).absolute())
    print(
        f'Structure: {structure}\n'
        f'Potential: {potential}\n'
        f'Directory: {directory}\n'
        f'Software: {software}\n'
        f'Calculation: {calculation}'
    )

    structure = get_structure(structure)
    potential = Potential.from_file(potential)
    if software == 'lammps':
        from ..io.lammps import modify_input_for_potential
        from lammps.sets import StaticSet, RelaxSet
        from lammps.inputs import LammpsData
        mapper = {
            'static': StaticSet,
            'relax': RelaxSet
        }
        input_set = mapper[calculation](LammpsData.from_structure(structure))
        modify_input_for_potential(input_set, potential)
        input_set.write_input(directory)



import json
import hashlib
import datetime as dt


def write_potential(dbm, potential):
    potential_str = json.dumps(potential.as_dict(with_parameters=False), sort_keys=True)
    potential_hash = hashlib.md5(potential_str.encode('utf-8')).hexdigest()

    with dbm.connection:
        result = dbm.connection.execute(
            'SELECT id FROM potential WHERE potential.id = ?',
            (potential_hash,)).fetchone()
        if result:
            return result['id']

        cursor = dbm.connection.execute(
            'INSERT INTO potential (id, schema) VALUES (?, ?)',
            (potential_hash, potential_str)
        )
        return cursor.lastrowid


def write_run_initial(dbm, potential):
    potential_id = write_potential(dbm, potential)

    with dbm.connection:
        cursor = dbm.connection.execute('''
        INSERT INTO run (potential_id, start_time, initial_parameters, indicies, bounds)
        VALUES (?, ?, ?, ?, ?)
        ''', (
            potential_id,
            dt.datetime.now(),
            potential.parameters.tolist(),
            potential.optimization_parameter_indicies.tolist(),
            potential.optimization_bounds.tolist()
        ))
        return potential_id, cursor.lastrowid


def write_run_final(dbm, run_id):
    with dbm.connection:
        cursor = dbm.connection.execute('UPDATE run SET end_time = ? WHERE id = ? AND end_time IS NULL', (dt.datetime.now(), run_id))


def write_evaluation(dbm, run_id, potential, result):
    with dbm.connection:
        errors = (
            result['parts']['forces'],
            result['parts']['stress'],
            result['parts']['energy'])
        if 'weights' in result:
            weights = (
                result['weights']['forces'],
                result['weights']['stress'],
                result['weights']['energy']
            )
        else:
            weights = (0, 0, 0)
        cursor = dbm.connection.execute('''
        INSERT INTO evaluation (run_id, parameters, sq_force_error, sq_stress_error, sq_energy_error, w_f, w_s, w_e)
        VALUES (?, ?, ?, ?, ?, ?, ?, ?)
        ''', (run_id, potential.optimization_parameters.tolist(),
              *errors, *weights))



from .potential import Potential
from .training import Training
from .config import Configuration
from .optimize import Optimize


def dftfit(configuration_schema, potential_schema, training_schema):
    configuration = Configuration(configuration_schema)
    potential = Potential(potential_schema)
    training = Training(training_schema, **configuration.training_kwargs)
    optimize = Optimize(
        dft_calculations=training.calculations,
        potential=potential,
        dbm=configuration.dbm,
        algorithm=configuration.algorithm,
        algorithm_kwargs=configuration.algorithm_kwargs,
        problem_kwargs=configuration.problem_kwargs,
    )
    population = optimize.population(configuration.population, seed=configuration.seed)
    optimize.optimize(population, steps=configuration.steps, seed=configuration.seed)



import click

import yaml
import json

from . import cli
from ..dftfit import dftfit


def load_filename(filename):
    if filename.endswith('.yaml') or filename.endswith('.yml'):
        return yaml.load(open(filename))
    elif filename.endswith('.json'):
        return json.load(open(filename))
    else:
        raise ValueError('only json and yaml files supported')


@cli.command()
@click.option('-t', '--training', type=click.Path(exists=True), required=True)
@click.option('-p', '--potential', type=click.Path(exists=True), required=True)
@click.option('-c', '--config', type=click.Path(exists=True))
@click.pass_context
def train(ctx, training, potential, config):
    dftfit(training_schema=load_filename(training),
           potential_schema=load_filename(potential),
           configuration_schema=load_filename(config))



import sqlite3 as sqlite
import json
import datetime as dt


POTENTIAL_TABLE = """
CREATE TABLE IF NOT EXISTS potential (
    id     TEXT PRIMARY KEY,
    schema TEXT NOT NULL
)
"""


RUN_TABLE = """
CREATE TABLE IF NOT EXISTS run (
    id                 INTEGER PRIMARY KEY,
    potential_id       TEXT NOT NULL,
    start_time         DATETIME NOT NULL,
    end_time           DATETIME,
    initial_parameters JSON NOT NULL,
    indicies           JSON NOT NULL,
    bounds             JSON NOT NULL,

    FOREIGN KEY(potential_id) REFERENCES potential(id)
)
"""

EVALUATION_TABLE = """
CREATE TABLE IF NOT EXISTS evaluation (
    id                 INTEGER PRIMARY KEY,
    run_id             INTEGER NOT NULL,
    parameters         JSON NOT NULL,
    sq_force_error     REAL NOT NULL,
    sq_stress_error    REAL NOT NULL,
    sq_energy_error    REAL NOT NULL,
    w_f                REAL NOT NULL,
    w_s                REAL NOT NULL,
    w_e                REAL NOT NULL,

    FOREIGN KEY(run_id) REFERENCES run(id)
)
"""


class DatabaseManager:
    def __init__(self, filename=None):
        self._connection = sqlite.connect(filename or ':memory:', detect_types=sqlite.PARSE_DECLTYPES)
        self._connection.row_factory = sqlite.Row
        self.register_types()
        self.create_tables()

    @staticmethod
    def adapt_json(d):
        return (json.dumps(d)).encode()

    @staticmethod
    def convert_json(data):
        return json.loads(data.decode())

    @staticmethod
    def adapt_datetime(datetime):
        return (datetime.strftime('%Y-%m-%d %H:%M:%S')).encode()

    @staticmethod
    def convert_datetime(data):
        return dt.datetime.strptime(data.decode(), '%Y-%m-%d %H:%M:%S')

    def register_types(self):
        sqlite.register_adapter(dt.datetime, self.adapt_datetime)
        sqlite.register_adapter(dict, self.adapt_json)
        sqlite.register_adapter(list, self.adapt_json)
        sqlite.register_adapter(tuple, self.adapt_json)
        sqlite.register_converter('datetime', self.convert_datetime)
        sqlite.register_converter('json', self.convert_json)

    def create_tables(self):
        self.connection.execute(POTENTIAL_TABLE)
        self.connection.execute(RUN_TABLE)
        self.connection.execute(EVALUATION_TABLE)

    @property
    def connection(self):
        return self._connection
